/*
 *                This source code is part of
 *
 *                          HelFEM
 *                             -
 * Finite element methods for electronic structure calculations on small systems
 *
 * Written by Susi Lehtola, 2018-
 * Copyright (c) 2018- Susi Lehtola
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 */
#ifndef SCF_HELPERS_H
#define SCF_HELPERS_H
#include <armadillo>
#include <helfem.h>

namespace helfem {
  namespace scf {
    /// Form density matrix
    arma::mat form_density(const arma::mat & C, size_t nocc);
    /// Enforce occupation of wanted symmetries
    void enforce_occupations(arma::mat & C, arma::vec & E, const arma::mat & S, const arma::ivec & nocc, const std::vector<arma::uvec> & m_idx);

    /// Enforce wanted symmetry in the Fock matrix (zero out off-diagonal blocks)
    arma::mat enforce_fock_symmetry(const arma::mat & Fin, const std::vector<arma::uvec> & m_idx);
    /// Average out the Fock matrix
    arma::mat fock_symmetry_average(const arma::mat & Fin, const std::vector< std::vector<arma::uvec> > & sym_idx);

    /// Solve generalized eigenvalue problem
    void eig_gsym(arma::vec & E, arma::mat & C, const arma::mat & F, const arma::mat & Sinvh);
    /// Solve generalized eigenvalue problem in subspaces
    void eig_gsym_sub(arma::vec & E, arma::mat & C, const arma::mat & F, const arma::mat & Sinvh, const std::vector<arma::uvec> & m_idx, bool verbose=true);
    /// Solve eigenvalue problem in subspaces
    void eig_sym_sub(arma::vec & E, arma::mat & C, const arma::mat & F, const std::vector<arma::uvec> & m_idx);

    /// Solve eigenvalue problem in subspace
    void eig_sub_wrk(arma::vec & E, arma::mat & Cocc, arma::mat & Cvirt, const arma::mat & F, size_t Nact);
    /// Sort eigenvectors
    void sort_eig(arma::vec & Eorb, arma::mat & Cocc, arma::mat & Cvirt, const arma::mat & Fao, size_t Nact, int maxit, double convthr);
    /// Solve subspace eigenproblem
    void eig_sub(arma::vec & E, arma::mat & Cocc, arma::mat & Cvirt, const arma::mat & F, size_t nsub, int maxit, double convthr);

    /// Iterative eigenvalue solver
    void eig_iter(arma::vec & E, arma::mat & Cocc, arma::mat & Cvirt, const arma::mat & F, const arma::mat & Sinvh, size_t nocc, size_t neig, size_t nsub, int maxit, double convthr);

    /// Random perturbation
    arma::mat perturbation_matrix(size_t N, double ampl);

    /// Form natural orbitals
    void form_NOs(const arma::mat & P, const arma::mat & Sh, const arma::mat & Sinvh, arma::mat & AO_to_NO, arma::mat & NO_to_AO, arma::vec & occs);
    /// Form half-inverse overlap
    inline arma::mat form_Sinvh(arma::mat S, bool chol=false) {
      return utils::invh(S, chol);
    }

    /// ROHF update to Fock matrices
    void ROHF_update(arma::mat & Fa_AO, arma::mat & Fb_AO, const arma::mat & P_AO, const arma::mat & Sh, const arma::mat & Sinvh, int nocca, int noccb);

    /// Human readable memory size
    std::string memory_size(size_t size);
    /// Parse number of alpha and beta electrons
    void parse_nela_nelb(int & nela, int & nelb, int & Q, int & M, int Z);

    /// Parse xc parameters
    arma::vec parse_xc_params(const std::string & input);
  }
}

#endif
